#define PATH_TO_SOUND "/root/compile/vsound/sounds"

#include "sabre_sound.h"
#include <getopt.h>


#define SERVER			"/tmp/serversocket"
#define MAXMSG			512
#define AUDIO_SPACE		100001 

int sock;
fd_set readfs;	/* data type represents file descriptor for select function,
                /* (actually it is a bit array) */
struct timeval timeout;
int use_alt_pcm_flag;
char audio_device[128];
char mixer_device[128];
char path_to_sound[1024];


int main(int argc, char **argv)
{
int a, b;
char message[MAXMSG];
struct sockaddr_un name;
size_t size;
int nbytes;
char filename[1024];
int speed;
int volume;
int offset;
int length;
int loop;
int played_flag;
FILE *fptr;
char *ptr;
int allow_interrupt;

/* kill any old socket */
fptr = popen("rm /tmp/serversocket", "r");
pclose(fptr);

/* default so we can start up */
/* the format is pathfilename:speed:volume:offset:length:loop */
/*
strcpy(message, "%s/wind_int.wav:0:100:0:100000:1", path_to_sound);
*/

/* defaults, else it wont startup */
strcpy(path_to_sound, PATH_TO_SOUND);

/*strcpy(filename, "/root/compile/vsound//wind_int.wav");*/
sprintf(filename, "%s/wind_int.wav", path_to_sound);
speed = 0;
volume = 100;
offset = 0;
length = 0;
loop = 0;
strcpy(audio_device, "/dev/dsp");
strcpy(mixer_device, "/dev/mixer");
use_alt_pcm_flag = 0;

/* command line flags */
while(1)
	{
	a = getopt(argc, argv, "ad:ehm:");
	if(a == -1) break;
	switch(a)
		{
		case 'a':
			use_alt_pcm_flag = 1;
			break;
		case 'd':
			strcpy(audio_device, optarg);
			break;
        case 'e': /* not to conflict with -d display option for x */
			debug_flag = 1;
			break;
		case 'h':
			fprintf(stdout,\
"Usage:\n\
sabre_sound [-a] [-d audio device] [-h] [-e] [-u presetfile] [-m mixer device]\n\n
\t-a  use pcm1 for volume control (use this when using /dev/dsp1).\n\
\t-d  audiodevice (/dev/dsp by default, or /dev/dsp1 etc..).\n\
\t-e  print debug messages (o.a. function calls and arguments).\n\
\t-h  help (this help).\n\
\t-m mixerdevice (/dev/mixer by default).\n\
");
			exit(0);	
			break;
		case '?':
			if (isprint(optopt) )
 				{
 				fprintf(stdout,\
 				"Unknown option `-%c' use -h for help.\n", optopt);
 				}
			else
				{
				fprintf(stdout,\
				"Unknown option character `\\x%x' use -h for help.\n",\
				optopt);
				}
			exit(1);
			break;			
		case 'm':
			strcpy(mixer_device, optarg);
			break;
		default:
			break;
		}/* end switch a */
	}/* end while getopt() */

/* Make the socket, then loop endlessly. */
sock = make_named_socket(SERVER);
while(1)
	{
	FD_ZERO(&readfs);
	FD_SET(sock, &readfs);

	timeout.tv_sec = 0;
	timeout.tv_usec = 0;
	b = select(FD_SETSIZE, &readfs, NULL, NULL, &timeout);
	if(b == -1)
		{
		fprintf(stdout, "select: returned < 0\n", "");
		}
	if( FD_ISSET(sock, &readfs) )
		{
		/* Wait for a datagram. */
		size = sizeof(name);
		nbytes = recvfrom(sock, message, MAXMSG, 0,
		(struct sockaddr *) & name, &size);
		if(nbytes < 0)
			{
			perror("recfrom (server)");
			exit(EXIT_FAILURE);
			}
		
		if(debug_flag)
			{
			fprintf(stdout,\
			"sabre_sound: main(): received message=%s\n",\
			message);
			}

		/* parse message */
		ptr = strtok(message, ":");/* filename */
		if(ptr) strcpy(filename, ptr);

		ptr = strtok(NULL, ":");/* speed */
		if(ptr) speed = atoi(ptr);

		ptr = strtok(NULL, ":");/* volume */
		if(ptr) volume = atoi(ptr);

		ptr = strtok(NULL, ":");/* offset */
		if(ptr) offset = atoi(ptr);
		
		ptr = strtok(NULL, ":");/* length */
		if(ptr) length = atoi(ptr);

		ptr = strtok(NULL, ":");/* loop */
		if(ptr) loop = atoi(ptr);

		played_flag = 0;
		}/* end if FD_SET */

/*
		FD_CLR(sock, &readfs);
*/

	if(debug_flag)
		{
		fprintf(stdout,\
		"main(): filename=%s speed=%d volume=%d offset=%d length=%d loop=%d\n",\
		filename, speed, volume, offset, length, loop);
		}	

	/* if not loop only play file once */
	if(loop == 0) if(played_flag) continue;

	/*
	in case of gun or cannnon i.e. no loop, do NOT allow interrupt,
	play file in full
	*/
	if(loop == 1) allow_interrupt = 1;
	else allow_interrupt = 0;
	
	if(! play(filename, speed, volume, offset, length, allow_interrupt) )
		{
		fprintf(stdout, "main(): play failed\n");
		exit(1);
		}
	played_flag = 1;
	
	}/* end while */
exit(0);
}/* end function main */


int play(\
char *filename, int speed, int volume, int offset, int length,\
int allow_interrupt)
{
struct passwd *userinfo;
int a, b, c, i, j, p, s, v;
char* local_filename;
int use_local_file_flag;
static int audionu;
char *abuf;
int sample_size;
static char *audiobuf;
static int slength;
double df;
double dv;
char *pos;
char *filebuf;
static int abuf_size;
static int audio_length;
int audio_rest;
unsigned long delay;
float qf, ql, qs;
char *ptr;
char speedstr[80];
char volumestr[80];
FILE *stream;
static int file_loaded_flag;
static char old_filename[1024];
static int old_speed;
static int old_volume;
static int old_offset;
static int old_length;
static int audio_open_flag;
static int mixer_fd;
static int mixer_open_flag;
static int sample_freq;
static int stereo;
static int bytes_per_sample;
static int bits_per_sample;
static char *audio_start;
static int current_speed;
static int current_length;

if(debug_flag)
	{
	fprintf(stdout,\
	"play(): arg filename=%s speed=%d volume=%d offset=%d length=%d\n\
	allow_interrupt=%d\n",\
	filename, speed, volume, offset, length,\
	allow_interrupt);
	}

/* argument check */
if(! filename) return(0);
if(offset < 0) return(0);
if(offset >= AUDIO_SPACE) return(0);
if(speed > 22000) speed = 22000;
if(length > (AUDIO_SPACE - offset) ) length = AUDIO_SPACE - offset;
if(volume < 0) return(0);
if(volume > 100) return(0);

if(debug_flag) fprintf(stdout, "play(): past argument check\n");

/* set current_speed if valid arg, else get from file header */
if(speed > 0) current_speed = speed;
else
	{
	if(file_loaded_flag) current_speed = sample_freq;
	}

/* set current_length if valid arg, else get from file header */
if(length > 0) current_length = length; 

if(! mixer_open_flag)
	{
	mixer_fd = open(mixer_device, O_RDWR, 0);
	if(mixer_fd < 0)
		{
		fprintf(stdout,\
		"play(): cannot open mixer device %s\n",\
		mixer_device);
		return(0);
		}
	mixer_open_flag = 1;
	}

if(! audio_open_flag)
	{
	audiobuf = malloc(AUDIO_SPACE);
	if(! audiobuf)
		{
		fprintf(stdout, "play(): cannot malloc audiobuf\n");
		return(0);
		}

	/* open audio device for play */
	audionu = open(audio_device, PLAY, 0);
	if(audionu == -1)
		{
		fprintf(stdout,\
		"play(): cannot open audio_device %s for playback\n",\
		audio_device);
		return(0);
		}

	/* get buffer size */
	ioctl(audionu, SNDCTL_DSP_GETBLKSIZE, &abuf_size);
	if(abuf_size == -1)
		{
		fprintf(stdout, "play(): cannot get blocksize\n");
		close(audionu);
		return(0);
		}

	audio_open_flag = 1;
	}/* end if ! audio_open_flag */
	
if( (! file_loaded_flag) || (strcmp(filename, old_filename) != 0))
	{
	if(! load_audio_file(\
	filename, audiobuf, &audio_start, &audio_length, &stereo,\
	&sample_freq, &bytes_per_sample, &bits_per_sample) )
		{
		fprintf(stdout, "play(): load_audio_file failed\n");
		close(audionu);
		return(0);
		}
	file_loaded_flag = 1;
	strcpy(old_filename, filename);
	ioctl(audionu, SNDCTL_DSP_RESET, &a);
	if(debug_flag)
		{
		fprintf(stdout, "play(): new filename now=%s\n", filename);
		}
	
	/* set sample size */
	sample_size = bits_per_sample;
	ioctl(audionu, SNDCTL_DSP_SAMPLESIZE, &sample_size);
	if(sample_size != bits_per_sample)
		{
		fprintf(stdout,\
		"play(): cannot set sample size to %d\n",\
		bits_per_sample);
		close(audionu);
		return(0);
		}
	if(debug_flag)
		{
		fprintf(stdout,\
		"play(): from file header: sample_size now=%d\n",\
		sample_size);
		}

	if(speed <= 0) current_speed = sample_freq;
	a = current_speed;
	ioctl(audionu, SNDCTL_DSP_SPEED, &a);
	old_speed = current_speed;
	if(debug_flag)
		{
		fprintf(stdout,\
		"play(): from file header: current_speed now=%d real=%d\n",\
		current_speed, a);
		}

	if(length <= 0) current_length = audio_length;

	}/* end if different audio file */

if(volume != old_volume)
	{
/*	ioctl(audionu, SNDCTL_DSP_RESET, &a);*/

	v = volume;

	if(use_alt_pcm_flag)
		{
		a = ioctl(mixer_fd, SOUND_MIXER_WRITE_ALTPCM, &v);
		}
	else
		{
		ioctl(mixer_fd, SOUND_MIXER_READ_PCM, &v);
		}
	old_volume = volume;
	}

if(offset != old_offset)
	{
	ioctl(audionu, SNDCTL_DSP_RESET, &a);
	old_offset = offset;
	}

if(debug_flag)
	{
	fprintf(stdout, "current_length=%d audio_length=%d offset=%d\n",\
	current_length, audio_length, offset);
	}

/* in case outside audio data, correct to keep it in */
if(current_length > (audio_length - offset) )
	{
	current_length = audio_length - offset;
	}
if(current_length != old_length)
	{
	ioctl(audionu, SNDCTL_DSP_RESET, &a);
	old_length = current_length;
	}
if(debug_flag) fprintf(stdout, "new current_length=%d\n", current_length);

if(current_speed != old_speed)
	{
	ioctl(audionu, SNDCTL_DSP_RESET, &a);
	s = current_speed;
	a = ioctl(audionu, SNDCTL_DSP_SPEED, &s);
/*
	if(s != current_speed)
		{
		fprintf(stdout, "play(): cannot set speed s=%d\n", s);
		close(audionu);
		return(0);
		}
*/
	old_speed = current_speed;
	if(debug_flag)
		{
		fprintf(stdout,\
		"play(): from arg: current_speed now=%d real=%d\n",\
		current_speed, s);
		}

	}/* end if current_speed != old_speed */

/* play the audio */
pos = audio_start + offset;
if(pos > (audio_start + current_length) )
	{
	fprintf(stdout, "play(): offset to big (outside file)\n");
	return(0);
	}
if( (offset + length) > current_length)
	{
	if(debug_flag)
		{
		fprintf(stdout, "play(): length to big (outside audio file)\n");
		}

	current_length -= offset;
/*	return(0);*/
	}

audio_rest = current_length;

slength = abuf_size;

if(debug_flag)
	{
	fprintf(stdout,\
	"play(): audio_start=%lu current_length=%d pos=%lu\n\
	audio_rest=%d slength=%d\n",\
	audio_start, current_length, pos, audio_rest, slength);
	}

while(1)
	{
	if(debug_flag)
		{
		a = (int) audio_start;
		b = (int) pos;
		fprintf(stdout, "position=%d\n", a - b);
		}

	/* modify if last piece */
	if(audio_rest < slength) slength = audio_rest;
	a = write(audionu, pos, slength);
	if(a != slength)
		{
		fprintf(stdout, "play(): error writing to audio device\n");
		close(audionu);
		return(0);
		}
	
	if(allow_interrupt)
		{
		/* be able to interrupt in the middle of a file */
		FD_ZERO(&readfs);
		FD_SET(sock, &readfs);

		timeout.tv_sec = 0;
		timeout.tv_usec = 0;
		b = select(FD_SETSIZE, &readfs, NULL, NULL, &timeout);
		if(b == -1)
			{
			fprintf(stdout, "select: returned < 0\n", "");
			}
		if( FD_ISSET(sock, &readfs) )
			{
			if(debug_flag)
				{
				fprintf(stdout, "****interrupting****\n");
				}
			return(1);
			}
		}/* end if allow interrupt */
	
	pos += slength;
	audio_rest -= slength;
	if( audio_rest <= 0) break;
	}/* end while */

/* close audio stream */
/*close(audionu);*/
usleep(500000);

if(debug_flag) fprintf(stdout, "play(): ready\n");
return(1);
}/* end function play */


int load_audio_file(\
char *filename, char *buffer, char **audio_start, int *length, int *stereo,\
int *sample_freq, int *bytes_per_sample, int *bits_per_sample)
{
int a, c, i;
struct stat *statptr;
char *space, *spaceptr;
FILE *fileptr;
size_t size;
char *ptr1, *ptr2, *ptr3, *ptr4;
/*static wave_header head;*/
static struct wave_header *head;

if(debug_flag)
	{
	fprintf(stdout, "load_audio_file arg filename=%s\n", filename);
	}

if(! filename) return(0);
if(! buffer) return(0);

head = (struct wave_header*)buffer;

if(debug_flag)
	{
	fprintf(stdout, "buffer=%lu head=%lu\n", buffer, head);
	}
	
*length = 0;

fileptr = fopen(filename, "r");
if(! fileptr)
	{
	if(debug_flag)
		{
		fprintf(stdout,\
		"load_audio_file(): cannot open file %s for read\n",\
		filename);
		}
	return(0);
	}/* end if open fails */

statptr = (struct stat*) malloc(sizeof(struct stat) );
if(! statptr) return(0);

fstat(fileno(fileptr), statptr);
size = statptr -> st_size;
if(size == 0)
	{
	free(statptr);
	return(0);
	}
*length = statptr -> st_size - sizeof(head);

/*
space = malloc(statptr -> st_size + 1);
free(statptr);
if(! space)
	{
	if(debug_flag)
		{
		fprintf(stdout,\
		"load_audio_file():  malloc failed\n");
		}
	return(0);
	}
*/

/*spaceptr = space;*/
spaceptr = buffer;
while(1)
	{
	while(1)/* error re read */
		{
		c = getc(fileptr);
		if(! ferror(fileptr) ) break;
		perror(" load_audio_file():  read failed ");
		}/* end while error re read */	

	if(feof(fileptr) )
		{
		fclose(fileptr);
		break;
		}
	*spaceptr = c;
	spaceptr++;
	}/* end while all lines from article body */
/* may not be there */
*spaceptr = 0;/* string termination */

*audio_start = buffer + sizeof(struct wave_header);

if(debug_flag)
	{
	fprintf(stdout, "buffer=%lu audio_start=%u\n", buffer, *audio_start);
	}

*sample_freq = head -> sample_fq;
*length = head -> data_length;
if(head -> modus == 2) *stereo = 1;
else *stereo = 0;
*bytes_per_sample = head -> byte_p_spl;
*bits_per_sample = head -> bit_p_spl;

if(debug_flag)
	{
	fprintf(stdout, "main_chunk=%s\n", head -> main_chunk);
	fprintf(stdout, "chunk_type=%s\n", head -> chunk_type);
	fprintf(stdout, "sub_chunk=%s\n", head -> sub_chunk);
	fprintf(stdout, "lenght_chunk=%d\n", head -> length_chunk);
	fprintf(stdout, "format=%d\n", head -> format);
	fprintf(stdout, "modus=%d\n", head -> modus);
	fprintf(stdout, "sample_fq=%d\n", head -> sample_fq);
	fprintf(stdout, "byte_p_sec=%d\n", head -> byte_p_sec);
	fprintf(stdout, "byte_p_spl=%d\n", head -> byte_p_spl);
	fprintf(stdout, "bit_p_spl=%d\n", head -> bit_p_spl);
	fprintf(stdout, "data_chunk=%s\n", head -> data_chunk);
	fprintf(stdout, "data_lenght=%lu\n", head -> data_length);
	}

return(1);
}/* end function load_audio_file */


int make_named_socket(const char *filename)
{
struct sockaddr_un name;
int sock;
size_t size;
 
/* Create the socket. */

sock = socket(PF_UNIX, SOCK_DGRAM, 0);
if(sock < 0)
	 {
	perror("socket");
	exit(EXIT_FAILURE);
	}

/* Bind a name to the socket. */

name.sun_family = AF_UNIX;
strcpy(name.sun_path, filename);

/*
The size of the address is
the offset of the start of the filename,
plus its length,
plus one for the terminating null byte. 
*/

size = offsetof(struct sockaddr_un, sun_path) + strlen(name.sun_path) + 1;

if(bind(sock, (struct sockaddr *) &name, size) < 0)
	{
	perror("bind");
	exit(EXIT_FAILURE);
	}

return sock;
}/* end function make_named_socket */


